"""
Websocket server.
"""

import asyncio
import json
import logging
from typing import Optional

from fastapi import APIRouter, FastAPI, WebSocket, WebSocketDisconnect
from pydantic import ValidationError
from pytest_relay_ws.observer import AnyMessage

router = APIRouter()

src_clients: set[WebSocket] = set()  # pytest-source clients
snk_clients: set[WebSocket] = set()  # pytest-sink clients

logger = logging.getLogger("uvicorn")


def _try_parse(data: str, ws: WebSocket) -> None:
    # ignore multiple identification attempts
    if ws in src_clients or ws in snk_clients:
        return

    msg: Optional[AnyMessage] = None
    try:
        msg = AnyMessage.model_validate_json(data, strict=False)
    except ValidationError as _:  # pylint: disable=broad-except,bare-except
        pass

    if msg is not None:
        match msg.type:
            case "pytest-source":
                logger.info("client %s added as pytest-source", ws.client)
                src_clients.add(ws)
            case "pytest-sink":
                logger.info("client %s added as pytest-sink", ws.client)
                snk_clients.add(ws)
            case _:
                pass


async def _try_send(data: str, ws: WebSocket) -> None:
    try:
        await ws.send_text(data)
    except Exception:  # pylint: disable=broad-except
        # sending may fail if the client disconnected while we're trying to relay.
        pass


async def _try_relay(data: str, source: WebSocket) -> None:
    # relay source messages to sinks and vice versa.
    # snapshot as lists since the sets may be manipulated concurrently.
    dst = None
    if source in src_clients:
        dst = list(snk_clients)
    elif source in snk_clients:
        dst = list(src_clients)
    else:
        return

    if not dst:
        return

    logger.info(
        "relaying"
        f"\n  {json.dumps(json.loads(data), indent=2)}\n"
        f"  to {[ws.client for ws in dst]}"
    )

    coros = [_try_send(data, ws) for ws in dst]
    if coros:
        # run all relays concurrently, slow clients won't block others
        await asyncio.gather(*coros, return_exceptions=True)


async def publish(data: str, to_sink: bool = False, to_source: bool = False) -> None:
    """
    Publishes a message to sinks and sources.
    """
    dst = []
    if to_sink:
        dst.extend(list(snk_clients))
    if to_source:
        dst.extend(list(src_clients))

    if not dst:
        return

    logger.info(
        "publishing"
        f"\n  {json.dumps(json.loads(data), indent=2)}\n"
        f"  to {[ws.client for ws in dst]}"
    )

    coros = [_try_send(data, ws) for ws in dst]
    if coros:
        # run all relays concurrently, slow clients won't block others
        await asyncio.gather(*coros, return_exceptions=True)


@router.websocket("/")
async def websocket_endpoint(websocket: WebSocket) -> None:
    """
    Endpoint for websocket connections.
    """
    await websocket.accept()
    # acknowledge the connection
    await _try_send(AnyMessage(type="pytest-conn-ack").model_dump_json(), websocket)
    try:
        while True:
            data = await websocket.receive_text()
            _try_parse(data, websocket)
            await _try_relay(data, websocket)
    except WebSocketDisconnect:
        src_clients.discard(websocket)
        snk_clients.discard(websocket)


# keep the 'app' definition at the end since all included routes must exist when defining the app
app = FastAPI(title="pytest-relay-run-ws")
app.include_router(router)



"""
Job execution API.
"""

import shlex
from http import HTTPStatus
from typing import List, Optional, Tuple

from fastapi import APIRouter, FastAPI, HTTPException
from fastapi.requests import Request
from fastapi.responses import Response

from pytest_relay_run.api.job import Job, JobManager
from pytest_relay_run.api.model import JobAction, JobActionRequest, JobRequest, JobResponse

router = APIRouter()
# router.include_router(items.router)
# router.include_router(others.router)

jobs = JobManager(max_sessions=100)


def _job_get(id_: str) -> Job:
    job: Optional[Job] = jobs.get(id_, None)
    if job is None:
        raise HTTPException(404, f"Job with ID '{id_}' found")
    return job


def _split_args(args: List[str]) -> List[str]:
    """
    Sanitize a list of arguments to be suitable for subprocess calls by splitting any arguments
    containing spaces into separate elements.
    """
    sanitized: List[str] = []

    for arg in args:
        sanitized.extend(shlex.split(arg))
    return sanitized


def _extract_session_id(args: List[str]) -> Optional[str]:
    """
    Extract the value of --session-id from a list of arguments.
    Returns the first session ID if provided, otherwise None.
    """
    it = iter(args)
    for arg in it:
        if arg.startswith("--session-id="):
            return arg.split("=", 1)[1]
        if arg == "--session-id":
            # Next element is the value, if it exists
            try:
                return next(it)
            except StopIteration:
                return None
    return None


def _parse_args(args: List[str], job_id: Optional[str]) -> Tuple[List[str], Optional[str]]:
    """
    Sanitizes the argument list and extracts the session ID, if available.
    Returns a tuple (args, job_id) where job_id is either the provided job_id or the session ID
    from the job's argument list.
    """
    args = [arg for arg in _split_args(args) if arg != "--collect-only"]
    session_id = _extract_session_id(args)

    if job_id is not None and session_id is not None and job_id != session_id:
        raise HTTPException(
            status_code=HTTPStatus.BAD_REQUEST,
            detail=f"The provided job ID '{job_id}' "
            f"does not match the session ID '{session_id}' "
            "provided in the job's argument list.",
        )

    job_id = job_id if job_id is not None else session_id
    return args, job_id


@router.post("/jobs", response_model=JobResponse, status_code=HTTPStatus.CREATED)
def job_create(payload: JobRequest, response: Response, request: Request) -> JobResponse:
    """
    Creates a new job.

    Notice that the job's ID is assigned to the test session, if it exists. Therefore, the
    relay runtime argument --session-id is not allowed as argument (or must be equivalent with
    the provided job ID).
    """
    if payload.id is not None and payload.id in jobs:
        raise HTTPException(
            status_code=HTTPStatus.CONFLICT, detail=f"Job with ID {payload.id} already exists"
        )

    args, job_id = _parse_args(payload.args, payload.id)
    job: Job = jobs.create_job(args, id_=job_id, collect_only=payload.collect_only)
    response.headers["Location"] = str(request.url_for("job_get", id_=job.id))
    return job.to_model()


@router.get("/jobs/{id_}", response_model=JobResponse)
def job_get(id_: str) -> JobResponse:
    """
    Retrieves the job with the given ID.
    """
    job = _job_get(id_)
    return job.to_model()


@router.get("/jobs", response_model=List[JobResponse])
def jobs_get() -> List[JobResponse]:
    """
    Retrieves all jobs.
    """
    return [job.to_model() for job in jobs.jobs()]


@router.post("/jobs/{id_}/actions", response_model=JobResponse)
def job_action(id_: str, req: JobActionRequest) -> JobResponse:
    """
    Triggers an action for the given job.
    """
    job = _job_get(id_)

    match req.action:
        case JobAction.START:
            job.start()
        case JobAction.STOP:
            job.stop()
        case _:
            raise HTTPException(HTTPStatus.BAD_REQUEST, f"Unsupported action '{req.action}'")

    # return the new state so the client can see the effect
    return job.to_model()


@router.delete("/jobs/{id_}", status_code=HTTPStatus.NO_CONTENT)
def job_delete(id_: str) -> None:
    """
    Deletes the job with the given ID. Notice that this does not terminate a running job, it is
    just an endpoint that can be used to explicitly clean up old results.
    """
    job = _job_get(id_)
    del jobs[job.id]


@router.delete("/jobs", status_code=HTTPStatus.NO_CONTENT)
def jobs_delete() -> None:
    """
    Deletes all jobs.
    """
    jobs.clear()


# keep the 'app' definition at the end since all included routes must exist when defining the app
app = FastAPI(title="pytest-relay-run-api")
app.include_router(router)



"""
Module for managing jobs (test executions).
"""

import asyncio
import signal
import subprocess
import threading
import uuid
from collections.abc import MutableMapping
from threading import Lock
from typing import Dict, Iterator, List, Optional

from pytest_relay_ws.observer import AnyMessage

from pytest_relay_run.api.model import JobResponse, JobState
from pytest_relay_run.ws.app import publish

# asynchronous loop to prevent "async" propagation for simple synchronous actions
_async_loop: Optional[asyncio.AbstractEventLoop] = None


def inject_async_loop(loop: asyncio.AbstractEventLoop) -> None:
    """
    Provides the running loop as reference
    """
    global _async_loop  # pylint: disable=global-statement
    _async_loop = loop


class Job:
    """
    Job (test session execution) instance.
    """

    id: str
    args: List[str]
    state: JobState
    returncode: Optional[int] = None
    text: Optional[str] = None
    _process: Optional[subprocess.Popen] = None

    def __init__(
        self, args: List[str], id_: Optional[str] = None, collect_only: bool = False
    ) -> None:
        self.id = str(uuid.uuid4()) if id_ is None else id_
        self.args = args
        self.state = JobState.CREATED

        if collect_only:
            self._collect()
        else:
            self._publish_state()

    def _publish_state(self) -> None:
        if _async_loop is None:
            raise RuntimeError("Async event loop not available")

        msg = AnyMessage(
            type="job",
            payload=self.to_model().model_dump(),
        )
        # "await" bridge: schedule the coroutine on the main loop
        asyncio.run_coroutine_threadsafe(publish(msg.model_dump_json(), to_sink=True), _async_loop)

    def _collect(self) -> None:
        proc: subprocess.Popen[str] = subprocess.Popen(  # pylint: disable=consider-using-with
            ["pytest", "--session-id", self.id, "--collect-only"] + self.args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,  # merge stderr into stdout
            text=True,
        )

        out, _err = proc.communicate()
        self.text = out
        self.returncode = proc.returncode
        self.state = JobState.COLLECTED
        self._publish_state()

    def start(self) -> None:
        """
        Starts a the pytest execution. The job is executed within a subprocess and monitored
        by a separate thread for completion.
        """
        if self.state not in (JobState.CREATED, JobState.COLLECTED):
            return

        self.state = JobState.IN_PROGRESS
        self._publish_state()

        self._process = subprocess.Popen(  # pylint: disable=consider-using-with
            ["pytest", "--session-id", self.id] + self.args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,  # merge stderr into stdout
            text=True,
            encoding="utf-8",
        )

        # monitor the session's state in background
        thread: threading.Thread = threading.Thread(target=self._monitor, daemon=True)
        thread.start()

    def _monitor(self) -> None:
        assert self._process

        out, _err = self._process.communicate()
        self.text = out
        self.returncode = self._process.returncode
        self.state = JobState.DONE
        self._publish_state()

    def stop(self) -> None:
        """
        Stops the execution of the job's process by sending SIGINT ("graceful" termination,
        equivalent to a keyboard interrupt). If SIGINT doesn't do the trick, the process is killed
        after a timeout.
        """
        if not self._process or self._process.poll() is not None:
            return

        self.state = JobState.TERMINATING
        self._publish_state()

        self._process.send_signal(signal.SIGINT)
        try:
            self._process.wait(timeout=1)
        except subprocess.TimeoutExpired:
            self._process.kill()

    def to_model(self) -> JobResponse:
        """
        Creates a response model that is used by the API for job information.
        """
        return JobResponse(
            id=self.id,
            state=self.state,
            returncode=self.returncode,
            text=self.text,
        )


class JobManager(MutableMapping[str, Job]):
    """
    Helper class for managing job instances.
    """

    def __init__(self, max_sessions: int = 20) -> None:
        self._jobs: Dict[str, Job] = {}
        self._max_sessions: int = max_sessions
        self._lock = Lock()

    def __getitem__(self, key: str) -> Job:
        with self._lock:
            return self._jobs[key]

    def __setitem__(self, key: str, value: Job) -> None:
        with self._lock:
            if len(self._jobs) >= self._max_sessions and key not in self._jobs:
                # Evict oldest
                oldest = next(iter(self._jobs))
                del self._jobs[oldest]
            self._jobs[key] = value

    def __delitem__(self, key: str) -> None:
        with self._lock:
            del self._jobs[key]

    def __iter__(self) -> Iterator[str]:
        with self._lock:
            return iter(self._jobs.copy())  # safe iteration

    def __len__(self) -> int:
        with self._lock:
            return len(self._jobs)

    def create_job(
        self, args: List[str], id_: Optional[str] = None, collect_only: bool = False
    ) -> Job:
        """
        Creates a job for the given arguments and immediately starts its execution if
        `collect_only` isn't defined.
        """
        job = Job(id_=id_, args=args, collect_only=collect_only)
        self[job.id] = job
        if not collect_only:
            job.start()
        return job

    def jobs(self) -> List[Job]:
        """
        Returns a snapshot of all jobs.
        """
        with self._lock:
            return list(self._jobs.values())



"""
.
"""

import argparse
import asyncio
from contextlib import asynccontextmanager
from typing import AsyncGenerator

import uvicorn
from fastapi import FastAPI

from pytest_relay_run.api.app import app as app_api
from pytest_relay_run.api.job import inject_async_loop
from pytest_relay_run.ws.app import app as app_ws

# import importlib
# version = importlib.metadata.version("pytest_relay_run")


@asynccontextmanager
async def _lifespan(_: FastAPI) -> AsyncGenerator[None, None]:
    """
    https://fastapi.tiangolo.com/advanced/events/
    """
    inject_async_loop(asyncio.get_running_loop())
    yield


app = FastAPI(lifespan=_lifespan)
# mount REST API at /api
app.mount("/api", app_api)
# mount WebSocket server at root
app.mount("/", app_ws)


def run() -> None:
    """
    Entry point for CLI.
    """
    parser = argparse.ArgumentParser(description="pytest-relay-run server")
    parser.add_argument("--host", type=str, default="0.0.0.0", help="application host")
    parser.add_argument("--port", type=int, default=8000, help="application port")

    args = parser.parse_args()
    uvicorn.run("pytest_relay_run.main:app", host=args.host, port=args.port)


if __name__ == "__main__":
    run()



"""
API model.
"""

from enum import Enum
from typing import List, Optional

from pydantic import BaseModel


class JobState(str, Enum):
    """Execution state of a job."""

    CREATED = "created"
    COLLECTED = "collected"
    IN_PROGRESS = "in-progress"
    DONE = "done"
    TERMINATING = "terminating"


class JobRequest(BaseModel):
    """Parameters accepted for job creation."""

    # TODO: consider accepting an executable such that the test can be run as a module
    # TODO: or using tools like poetry, uv, poe ... it could even support `tox` to some extent.

    id: Optional[str]
    args: List[str]
    collect_only: bool = False


class JobResponse(BaseModel):
    """Job information exposed via the API."""

    id: str
    state: JobState
    returncode: Optional[int] = None
    text: Optional[str] = None


class JobAction(str, Enum):
    """Action identifiers supported for jobs."""

    START = "start"
    STOP = "stop"


class JobActionRequest(BaseModel):
    """Model for action requests."""

    action: JobAction
